#!/usr/bin/python3

## Copyright (C) 2008, 2009, 2010, 2012, 2014 Red Hat, Inc.
## Authors:
##  Tim Waugh <twaugh@redhat.com>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import config
import gettext
gettext.install(domain=config.PACKAGE, localedir=config.localedir)
import cupshelpers
import urllib.parse
import ppdippstr
import socket
from debug import *

class PhysicalDevice:
    def __init__(self, device):
        self.devices = None
        self._network_host = None
        self.dnssd_hostname = None
        self._cupsserver = False
        self.firsturi = None
        self.add_device (device)
        self._user_data = {}
        self._ppdippstr = ppdippstr.backends

    def _canonical_id (self, device):
        if hasattr (device, "id_dict"):
            mfg = device.id_dict.get ('MFG', '')
            mdl = device.id_dict.get ('MDL', '')

            if mfg == '' or mdl.lower ().startswith (mfg.lower ()):
                make_and_model = mdl
            else:
                make_and_model = "%s %s" % (mfg, mdl)
        else:
             make_and_model = device.make_and_model

        return cupshelpers.ppds.ppdMakeModelSplit (make_and_model)

    def _add_dot_local_if_needed(self, hostname):
        if (hostname is None):
            return None
        if ((not '.' in hostname) and (not ':' in hostname) and
            (hostname != 'localhost')):
            return hostname + '.local'
        else:
            return hostname

    def _get_address (self, hostname):
        try:
            address = socket.getaddrinfo(hostname, 0,
                                         family=socket.AF_INET)[0][4][0]
        except:
            try:
                address = socket.getaddrinfo(hostname, 0,
                                             family=socket.AF_INET6)[0][4][0]
            except:
                address = None
        return address

    def _get_host_from_uri (self, uri):
        hostport = None
        host = None
        dnssdhost = None
        (scheme, rest) = urllib.parse.splittype (uri)
        if scheme == 'hp' or scheme == 'hpfax':
            ipparam = None
            if rest.startswith ("/net/"):
                (rest, ipparam) = urllib.parse.splitquery (rest[5:])

            if ipparam is not None:
                if ipparam.startswith("ip="):
                    hostport = ipparam[3:]
                elif ipparam.startswith ("hostname="):
                    hostport = ipparam[9:]
                elif ipparam.startswith("zc="):
                    dnssdhost = ipparam[3:]
                else:
                    return None, None
            else:
                return None, None
        elif scheme == 'dnssd' or scheme == 'mdns':
            # The URIs of the CUPS "dnssd" backend do not contain the host
            # name of the printer
            return None, None
        else:
            (hostport, rest) = urllib.parse.splithost (rest)
            if hostport is None:
                return None, None

        if hostport:
            (host, port) = urllib.parse.splitport (hostport)

        if (host):
            ip = None
            try:
                ip = self._get_address(host)
                if ip:
                    host = ip
            except:
                pass
        elif (dnssdhost):
            try:
                host = self._get_address(dnssdhost)
            except:
                host = None

        return self._add_dot_local_if_needed(host), \
            self._add_dot_local_if_needed(dnssdhost)

    def add_device (self, device):
        if self._network_host or self.dnssd_hostname:
            host, dnssdhost = self._get_host_from_uri (device.uri)
            if (hasattr (device, 'address')):
                host = self._add_dot_local_if_needed(device.address)
            if (hasattr (device, 'hostname') and dnssdhost is None):
                dnssdhost = self._add_dot_local_if_needed(device.hostname)
            if (host is None and dnssdhost is None) or \
               not ((host and self._network_host and
                     host == self._network_host) or
                    (host and self.dnssd_hostname and
                     host == self.dnssd_hostname) or
                    (dnssdhost and self._network_host and
                     dnssdhost == self._network_host) or
                    (dnssdhost and self.dnssd_hostname and
                     dnssdhost == self.dnssd_hostname)) or \
               (host is None and self.dnssd_hostname is None) or \
               (dnssdhost is None and self._network_host is None):
                raise ValueError
        else:
            (mfg, mdl) = self._canonical_id (device)
            if self.devices is None:
                self.mfg = mfg
                self.mdl = mdl
                self.mfg_lower = mfg.lower ()
                self.mdl_lower = mdl.lower ()
                self.sn = device.id_dict.get ('SN', '')
                self.devices = []
            else:
                def nicest (a, b):
                    def count_lower (s):
                        l = s.lower ()
                        n = 0
                        for i in range (len (s)):
                            if l[i] != s[i]:
                                n += 1
                        return n
                    if count_lower (b) < count_lower (a):
                        return b
                    return a

                self.mfg = nicest (self.mfg, mfg)
                self.mdl = nicest (self.mdl, mdl)

                sn = device.id_dict.get ('SN', '')
                if sn != '' and self.sn != '' and sn != self.sn:
                    raise ValueError

        if device.type == "socket":
            # Remove default port to more easily find duplicate URIs
            device.uri = device.uri.replace (":9100", "")
        if (device.uri.startswith('ipp:') and \
            device.uri.find('/printers/') != -1) or \
           ((device.uri.startswith('dnssd:') or \
             device.uri.startswith('mdns:')) and \
            device.uri.endswith('/cups')):
            # CUPS server
            self._cupsserver = True
        elif self._cupsserver:
            # Non-CUPS queue on a CUPS server, drop this one
            return
        for d in self.devices:
            if d.uri == device.uri:
                return

        # Use the URI of the very first device added as a kind of identifier
        # for this physical device record, to make debugging easier
        if not self.firsturi:
            self.firsturi = device.uri;

        self.devices.append (device)
        self.devices.sort ()

        if (not self._network_host or not self.dnssd_hostname) and \
           device.device_class == "network":
            # We just added a network device.
            self._network_host, dnssdhost = \
                self._get_host_from_uri (device.uri)
            if dnssdhost:
                self.dnssd_hostname = dnssdhost;

        if (hasattr (device, 'address') and self._network_host is None):
            address = device.address
            if address:
                self._network_host = self._add_dot_local_if_needed(address)

        if (hasattr (device, 'hostname') and self.dnssd_hostname is None):
            hostname = device.hostname
            if hostname:
                self.dnssd_hostname = self._add_dot_local_if_needed(hostname)

        if (self.dnssd_hostname and self._network_host is None):
            try:
                self._network_host = self._get_address(hostname);
            except:
                self._network_host = None

        debugprint("Device %s added to physical device: %s" %
                   (device.uri, repr(self)))

    def get_devices (self):
        return self.devices

    def get_info (self):
        # If the manufacturer/model is not known, or useless (in the
        # case of the hpfax backend or a dnssd URI pointing to a remote
        # CUPS queue), show the device-info field instead.
        if (self.devices[0].uri.startswith('ipp:') and \
            self.devices[0].uri.find('/printers/') != -1) or \
           ((self.devices[0].uri.startswith('dnssd:') or \
             self.devices[0].uri.startswith('mdns:')) and \
            self.devices[0].uri.endswith('/cups')):
            if not self.dnssd_hostname:
                info = "%s" % self._network_host
            elif not self._network_host or self._network_host.find(":") != -1:
                info = "%s" % self.dnssd_hostname
            else:
                if self._network_host != self.dnssd_hostname:
                    info = "%s (%s)" % (self.dnssd_hostname, self._network_host)
                else:
                    info = "%s" % self._network_host
        elif self.mfg == '' or \
           (self.mfg == "HP" and self.mdl.startswith("Fax")):
            info = self._ppdippstr.get (self.devices[0].info)
        else:
            info = "%s %s" % (self.mfg, self.mdl)
        if ((self._network_host and len (self._network_host) > 0) or \
            (self.dnssd_hostname and len (self.dnssd_hostname) > 0)) and not \
            ((self.devices[0].uri.startswith('dnssd:') or \
              self.devices[0].uri.startswith('mdns:')) and \
              self.devices[0].uri.endswith('/cups')) and \
            (not self._network_host or \
             info.find(self._network_host) == -1) and \
            (not self.dnssd_hostname or \
             info.find(self.dnssd_hostname) == -1):
            if not self.dnssd_hostname:
                info += " (%s)" % self._network_host
            elif not self._network_host:
                info += " (%s)" % self.dnssd_hostname
            else:
                info += " (%s, %s)" % (self.dnssd_hostname, self._network_host)
        elif len (self.sn) > 0:
            info += " (%s)" % self.sn
        return info

    # User data
    def set_data (self, key, value):
        self._user_data[key] = value

    def get_data (self, key):
        return self._user_data.get (key)

    def __str__ (self):
        return "(description: %s)" % self.__repr__ ()

    def __repr__ (self):
        return ("<PhysicalDevice.PhysicalDevice (%s,%s,%s,%s,%s,%s)>" %
                (self.mfg, self.mdl, self.sn, self._network_host,
                 self.dnssd_hostname, self.firsturi))

    def __eq__(self, other):
        if type (other) != type (self):
            return False

        if not (((not self._network_host or len (self._network_host) == 0) and
                 (not other._network_host or len (other._network_host) == 0) and
                 (not self.dnssd_hostname or len (self.dnssd_hostname) == 0) and
                 (not other.dnssd_hostname or
                  len (other.dnssd_hostname) == 0)) or
                (self._network_host and len (self._network_host) > 0 and
                 other._network_host and len (other._network_host) > 0 and
                 self._network_host == other._network_host) or
                (self.dnssd_hostname and len (self.dnssd_hostname) > 0 and
                 other.dnssd_hostname and len (other.dnssd_hostname) > 0 and
                 self.dnssd_hostname == other.dnssd_hostname) or
                (self._network_host and len (self._network_host) > 0 and
                 other.dnssd_hostname and len (other.dnssd_hostname) > 0 and
                 self._network_host == other.dnssd_hostname) or
                (self.dnssd_hostname and len (self.dnssd_hostname) > 0 and
                 other._network_host and len (other._network_host) > 0 and
                 self.dnssd_hostname == other._network_host)):
            return False

        devs = other.get_devices()
        if devs:
            uris = [x.uri for x in self.devices]
            for dev in devs:
                if dev.uri in uris:
                    # URI match
                    return True

        if ((other.mfg == '' and other.mdl == '') or
            (self.mfg == ''  and self.mdl == '')):
            if other.mfg == '' and self.mfg == '':
                # Both just a backend, not a real physical device.
                return self.devices[0] == other.devices[0]

            # One or other is just a backend, not a real physical device.
            return False

        def split_make_and_model (dev):
            if dev.mfg == '' or dev.mdl.lower ().startswith (dev.mfg.lower ()):
                make_and_model = dev.mdl
            else:
                make_and_model = "%s %s" % (dev.mfg, dev.mdl)
            (mfg, mdl) = cupshelpers.ppds.ppdMakeModelSplit (make_and_model)
            return (cupshelpers.ppds.normalize (mfg),
                    cupshelpers.ppds.normalize (mdl))

        (our_mfg, our_mdl) = split_make_and_model (self)
        (other_mfg, other_mdl) = split_make_and_model (other)

        if our_mfg != other_mfg:
            return False

        if our_mfg == "hp" and self.sn != '' and self.sn == other.sn:
            return True

        if our_mdl != other_mdl:
            return False

        if self.sn == '' or other.sn == '':
            return True

        return self.sn == other.sn

    def __lt__(self, other):
        if type (other) != type (self):
            return False

        if self == other:
            return False;

        if self._network_host != other._network_host:
            if self._network_host is None:
                return True

            if other._network_host is None:
                return False

            return self._network_host < other._network_host

        if self.dnssd_hostname != other.dnssd_hostname:
            if self.dnssd_hostname is None:
                return True

            if other.dnssd_hostname is None:
                return False

            return self.dnssd_hostname < other.dnssd_hostname

        devs = other.get_devices()
        if devs:
            uris = [x.uri for x in self.devices]
            for dev in devs:
                if dev.uri in uris:
                    # URI match, so compare equal. Not less than.
                    return False

        if ((other.mfg == '' and other.mdl == '') or
            (self.mfg == ''  and self.mdl == '')):
            if other.mfg == '' and self.mfg == '':
                # Both just a backend, not a real physical device.
                return self.devices[0] < other.devices[0]

            # One or other is just a backend, not a real physical device.
            return other.mfg == '' and other.mdl == ''

        def split_make_and_model (dev):
            if dev.mfg == '' or dev.mdl.lower ().startswith (dev.mfg.lower ()):
                make_and_model = dev.mdl
            else:
                make_and_model = "%s %s" % (dev.mfg, dev.mdl)
            (mfg, mdl) = cupshelpers.ppds.ppdMakeModelSplit (make_and_model)
            return (cupshelpers.ppds.normalize (mfg),
                    cupshelpers.ppds.normalize (mdl))

        (our_mfg, our_mdl) = split_make_and_model (self)
        (other_mfg, other_mdl) = split_make_and_model (other)

        if our_mfg != other_mfg:
            return our_mfg < other_mfg

        if our_mdl != other_mdl:
            return our_mdl < other_mdl

        if self.sn == '' or other.sn == '':
            return False

        return self.sn < other.sn

if __name__ == '__main__':
    import authconn
    c = authconn.Connection ()
    devices = cupshelpers.getDevices (c)

    physicaldevices = []
    for device in devices.values ():
        physicaldevice = PhysicalDevice (device)
        try:
            i = physicaldevices.index (physicaldevice)
            physicaldevices[i].add_device (device)
        except ValueError:
            physicaldevices.append (physicaldevice)

    physicaldevices.sort ()
    for physicaldevice in physicaldevices:
        print(physicaldevice.get_info ())
        devices = physicaldevice.get_devices ()
        for device in devices:
            print(" ", device)



#!/usr/bin/python3

## system-config-printer

## Copyright (C) 2006, 2007, 2008, 2009, 2010 Red Hat, Inc.
## Copyright (C) 2006 Florian Festi <ffesti@redhat.com>
## Copyright (C) 2006, 2007, 2008, 2009 Tim Waugh <twaugh@redhat.com>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import fnmatch
import re
import xml.etree.ElementTree
from .cupshelpers import parseDeviceID

def PreferredDrivers (filename):
    preferreddrivers = xml.etree.ElementTree.XML (open (filename).read ())
    return list(preferreddrivers)

class DeviceIDMatch:
    """
    A set of fields and regular expressions for matching a Device ID.
    """
    def __init__ (self):
        self._re = dict()

    def add_field (self, field, pattern):
        self._re[field.upper ()] = re.compile (pattern, re.I)

    def match (self, deviceid):
        """
        Match against a parsed Device ID dictionary.

        The CMD field is treated specially.  If any of the
        comma-separated words in this field value match, the Device ID
        pattern is considered to match.
        """

        for field, match in self._re.items ():
            if field not in deviceid:
                return False

            if field == "CMD":
                this_field_matches = False
                for cmd in deviceid[field]:
                    if match.match (cmd):
                        this_field_matches = True
                        break

                if not this_field_matches:
                    return False

            if not match.match (deviceid[field]):
                return False

        return True

class DriverType:
    """
    A type of driver.
    """

    FIT_EXACT_CMD =     "exact-cmd"
    FIT_EXACT =         "exact"
    FIT_CLOSE =         "close"
    FIT_GENERIC =       "generic"
    FIT_NONE =          "none"

    def __init__ (self, name):
        self.name = name
        self.ppd_name = None
        self.attributes = []
        self.deviceid = []

        class AlwaysTrue:
            def get (self, k, d=None):
                return True

        self._fit = AlwaysTrue ()
        self._packagehint = None

    def add_ppd_name (self, pattern):
        """
        An optional PPD name regular expression.
        """
        self.ppd_name = re.compile (pattern, re.I)

        # If the PPD name pattern includes a scheme, we can perhaps
        # deduce which package would provide this driver type.
        if self._packagehint is not None:
            return

        parts = pattern.split (":", 1)
        if len (parts) > 1:
            scheme = parts[0]
            if scheme == "drv":
                rest = parts[1]
                if rest.startswith ("///"):
                    drv = rest[3:]
                    f = drv.rfind ("/")
                    if f != -1:
                        drv = drv[:f]
                        self._packagehint = "/usr/share/cups/drv/%s" % drv
            else:
                self._packagehint = "/usr/lib/cups/driver/%s" % scheme

    def add_attribute (self, name, pattern):
        """
        An optional IPP attribute name and regular expression to match
        against its values.
        """
        self.attributes.append ((name, re.compile (pattern, re.I)))

    def add_deviceid_match (self, deviceid_match):
        """
        An optional IEEE 1284 Device ID match.
        """
        self.deviceid.append (deviceid_match)

    def add_fit (self, text):
        self._fit = {}
        for fittype in text.split():
            self._fit[fittype] = True

            # <fit>exact</fit> matches exact-cmd as well
            if fittype == self.FIT_EXACT:
                self._fit[self.FIT_EXACT_CMD] = True

    def set_packagehint (self, hint):
        self._packagekit = hint

    def get_name (self):
        """
        Return the name for this driver type.
        """
        return self.name

    def __repr__ (self):
        return "<DriverType %s instance at 0x%x>" % (self.name, id (self))

    def match (self, ppd_name, attributes, fit):
        """
        Return True if there is a match for all specified criteria.

        ppdname: string

        attributes: dict

        fit: string
        """

        matches = self._fit.get (fit, False)
        if matches and self.ppd_name and not self.ppd_name.match (ppd_name):
            matches = False

        if matches:
            for name, match in self.attributes:
                if name not in attributes:
                    matches = False
                    break

                values = attributes[name]
                if not isinstance (values, list):
                    # In case getPPDs() was used instead of getPPDs2()
                    values = [values]

                any_value_matches = False
                for value in values:
                    if match.match (value):
                        any_value_matches = True
                        break

                if not any_value_matches:
                    matches = False
                    break

        if matches:
            if self.deviceid and "ppd-device-id" not in attributes:
                matches = False
            elif self.deviceid:
                # This is a match if any of the ppd-device-id values
                # match.
                deviceidlist = attributes["ppd-device-id"]
                if not isinstance (deviceidlist, list):
                    # In case getPPDs() was used instead of getPPDs2()
                    deviceidlist = [deviceidlist]

                any_id_matches = False
                for deviceidstr in deviceidlist:
                    deviceid = parseDeviceID (deviceidstr)
                    for match in self.deviceid:
                        if match.match (deviceid):
                            any_id_matches = True
                            break

                if not any_id_matches:
                    matches = False

        return matches

    def get_packagehint (self):
        return None

class DriverTypes:
    """
    A list of driver types.
    """

    def __init__ (self):
        self.drivertypes = []

    def load (self, drivertypes):
        """
        Load the list of driver types from an XML file.
        """

        types = []
        for drivertype in list(drivertypes):
            t = DriverType (drivertype.attrib["name"])

            for child in list(drivertype):
                if child.tag == "ppdname":
                    t.add_ppd_name (child.attrib["match"])
                elif child.tag == "attribute":
                    t.add_attribute (child.attrib["name"],
                                     child.attrib["match"])
                elif child.tag == "deviceid":
                    deviceid_match = DeviceIDMatch ()
                    for field in list(child):
                        if field.tag == "field":
                            deviceid_match.add_field (field.attrib["name"],
                                                      field.attrib["match"])

                    t.add_deviceid_match (deviceid_match)
                elif child.tag == "fit":
                    t.add_fit (child.text)

            types.append (t)

        self.drivertypes = types

    def match (self, ppdname, ppddict, fit):
        """
        Return the first matching drivertype for a PPD, given its name,
        attributes, and fitness, or None if there is no match.
        """

        for drivertype in self.drivertypes:
            if drivertype.match (ppdname, ppddict, fit):
                return drivertype

        return None

    def filter (self, pattern):
        """
        Return the subset of driver type names that match a glob
        pattern.
        """

        return fnmatch.filter ([x.get_name () for x in self.drivertypes],
                               pattern)

    def get_ordered_ppdnames (self, drivertypes, ppdsdict, fit):
        """
        Given a list of driver type names, a dict of PPD attributes by
        PPD name, and a dict of driver fitness status codes by PPD
        name, return a list of tuples in the form (driver-type-name,
        PPD-name), representing PPDs that match the list of driver
        types.

        The returned tuples will have driver types in the same order
        as the driver types given, with the exception that any
        blacklisted driver types will be omitted from the returned
        result.
        """

        ppdnames = []

        # First find out what driver types we have
        ppdtypes = {}
        fit_default = DriverType.FIT_CLOSE
        for ppd_name, ppd_dict in ppdsdict.items ():
            drivertype = self.match (ppd_name, ppd_dict, fit.get (ppd_name,
                                                                  fit_default))
            if drivertype:
                name = drivertype.get_name ()
            else:
                name = "none"

            m = ppdtypes.get (name, [])
            m.append (ppd_name)
            ppdtypes[name] = m

        # Now construct the list.
        for drivertypename in drivertypes:
            for ppd_name in ppdtypes.get (drivertypename, []):
                if ppd_name in ppdnames:
                    continue

                ppdnames.append ((drivertypename, ppd_name))

        return ppdnames

class PrinterType:
    """
    A make-and-model pattern and/or set of IEEE 1284 Device ID
    patterns for matching a set of printers, together with an ordered
    list of driver type names.
    """

    def __init__ (self):
        self.make_and_model = None
        self.deviceid = []
        self.drivertype_patterns = []
        self.avoid = set()
        self.blacklist = set()

    def add_make_and_model (self, pattern):
        """
        Set a make-and-model regular expression.  Only one is permitted.
        """
        self.make_and_model = re.compile (pattern, re.I)

    def add_deviceid_match (self, deviceid_match):
        """
        Add a Device ID match.
        """
        self.deviceid.append (deviceid_match)

    def add_drivertype_pattern (self, name):
        """
        Append a driver type pattern.
        """
        self.drivertype_patterns.append (name.strip ())

    def get_drivertype_patterns (self):
        """
        Return the list of driver type patterns.
        """
        return self.drivertype_patterns

    def add_avoidtype_pattern (self, name):
        """
        Add an avoid driver type pattern.
        """
        self.avoid.add (name)

    def get_avoidtype_patterns (self):
        """
        Return the set of driver type patterns to avoid.
        """
        return self.avoid

    def add_blacklisted (self, name):
        """
        Add a blacklisted driver type pattern.
        """
        self.blacklist.add (name)

    def get_blacklist (self):
        """
        Return the set of blacklisted driver type patterns.
        """
        return self.blacklist

    def match (self, make_and_model, deviceid):
        """
        Return True if there are no constraints to match against; if
        the make-and-model pattern matches; or if all of the IEEE 1284
        Device ID patterns match.

        The deviceid parameter must be a dict indexed by Device ID
        field key, of strings; except for the CMD field which must be
        a list of strings.

        Return False otherwise.
        """

        matches = (self.make_and_model is None and self.deviceid == [])
        if self.make_and_model:
            if self.make_and_model.match (make_and_model):
                matches = True

        if not matches:
            for match in self.deviceid:
                if match.match (deviceid):
                    matches = True
                    break

        return matches

class PreferenceOrder:
    """
    A policy for choosing the preference order for drivers.
    """

    def __init__ (self):
        self.ptypes = []

    def load (self, preferreddrivers):
        """
        Load the policy from an XML file.
        """

        for printer in list(preferreddrivers):
            ptype = PrinterType ()
            for child in list(printer):
                if child.tag == "make-and-model":
                    ptype.add_make_and_model (child.attrib["match"])
                elif child.tag == "deviceid":
                    deviceid_match = DeviceIDMatch ()
                    for field in list(child):
                        if field.tag == "field":
                            deviceid_match.add_field (field.attrib["name"],
                                                      field.attrib["match"])
                    ptype.add_deviceid_match (deviceid_match)

                elif child.tag == "drivers":
                    for drivertype in list(child):
                        ptype.add_drivertype_pattern (drivertype.text)

                elif child.tag == "avoid":
                    for drivertype in list(child):
                        ptype.add_avoidtype_pattern (drivertype.text)

                elif child.tag == "blacklist":
                    for drivertype in list(child):
                        ptype.add_blacklisted (drivertype.text)

            self.ptypes.append (ptype)

    def get_ordered_types (self, drivertypes, make_and_model, deviceid):
        """
        Return an accumulated list of driver types from all printer
        types that match a given printer's device-make-and-model and
        IEEE 1284 Device ID.

        The deviceid parameter must be None or a dict indexed by
        short-form upper-case field keys.
        """

        if deviceid is None:
            deviceid = {}

        if make_and_model is None:
            make_and_model = ""

        orderedtypes = []
        blacklist = set()
        avoidtypes = set()
        for ptype in self.ptypes:
            if ptype.match (make_and_model, deviceid):
                for pattern in ptype.get_drivertype_patterns ():
                    # Match against the glob pattern
                    for drivertype in drivertypes.filter (pattern):
                        # Add each result if not already in the list.
                        if drivertype not in orderedtypes:
                            orderedtypes.append (drivertype)

                for pattern in ptype.get_avoidtype_patterns ():
                    # Match against the glob pattern.
                    for drivertype in drivertypes.filter (pattern):
                        # Add each result to the set.
                        avoidtypes.add (drivertype)

                for pattern in ptype.get_blacklist ():
                    # Match against the glob pattern.
                    for drivertype in drivertypes.filter (pattern):
                        # Add each result to the set.
                        blacklist.add (drivertype)

        if avoidtypes:
            avoided = []
            for t in avoidtypes:
                try:
                    i = orderedtypes.index (t)
                    del orderedtypes[i]
                    avoided.append (t)
                except IndexError:
                    continue

            orderedtypes.extend (avoided)

        if blacklist:
            # Remove blacklisted drivers.
            remaining = []
            for t in orderedtypes:
                if t not in blacklist:
                    remaining.append (t)

            orderedtypes = remaining

        return orderedtypes


def test (xml_path=None, attached=False, deviceid=None, debug=False):
    import cups
    import locale
    from . import ppds
    from pprint import pprint
    from time import time
    import os.path

    if debug:
        def debugprint (x):
            print (x)

        ppds.set_debugprint_fn (debugprint)
            
    locale.setlocale (locale.LC_ALL, "")
    if xml_path is None:
        xml_path = os.path.join (os.path.join (os.path.dirname (__file__),
                                               ".."),
                                 "xml")

    os.environ["CUPSHELPERS_XMLDIR"] = xml_path
    xml_path = os.path.join (xml_path, "preferreddrivers.xml")
    loadstart = time ()
    (xmldrivertypes, xmlpreferenceorder) = PreferredDrivers (xml_path)
    drivertypes = DriverTypes ()
    drivertypes.load (xmldrivertypes)

    preforder = PreferenceOrder ()
    preforder.load (xmlpreferenceorder)
    loadtime = time () - loadstart
    if debug:
        print("Time to load %s: %.3fs" % (xml_path, loadtime))

    c = cups.Connection ()
    try:
        cupsppds = c.getPPDs2 ()
    except AttributeError:
        # getPPDs2 requires pycups >= 1.9.52 
        cupsppds = c.getPPDs ()

    ppdfinder = ppds.PPDs (cupsppds)

    if attached or deviceid:
        if attached:
            cups.setUser ("root")
            devices = c.getDevices ()
        else:
            devid = parseDeviceID (deviceid)
            devices = { "xxx://yyy":
                            { "device-id": deviceid,
                              "device-make-and-model": "%s %s" % (devid["MFG"],
                                                                  devid["MDL"])
                              }
                        }

        for uri, device in devices.items ():
            if uri.find (":") == -1:
                continue

            devid = device.get ("device-id", "")
            if isinstance (devid, list):
                devid = devid[0]

            if not devid:
                continue

            if not uri.startswith ("xxx:"):
                print (uri)

            id_dict = parseDeviceID (devid)
            fit = ppdfinder.getPPDNamesFromDeviceID (id_dict["MFG"],
                                                     id_dict["MDL"],
                                                     id_dict["DES"],
                                                     id_dict["CMD"],
                                                     uri)

            mm = device.get ("device-make-and-model", "")
            orderedtypes = preforder.get_ordered_types (drivertypes,
                                                        mm, id_dict)

            ppds = {}
            for ppdname in fit.keys ():
                ppds[ppdname] = ppdfinder.getInfoFromPPDName (ppdname)

            orderedppds = drivertypes.get_ordered_ppdnames (orderedtypes,
                                                            ppds,
                                                            fit)
            i = 1
            for t, ppd in orderedppds:
                print("%d  %s\n    (%s, %s)" % (i, ppd, t, fit[ppd]))
                i += 1
    else:
        for make in ppdfinder.getMakes ():
            for model in ppdfinder.getModels (make):
                ppdsdict = ppdfinder.getInfoFromModel (make, model)
                mm = make + " " + model
                orderedtypes = preforder.get_ordered_types (drivertypes,
                                                            mm, None)

                fit = {}
                for ppdname in ppdsdict.keys ():
                    fit[ppdname] = DriverType.FIT_CLOSE

                orderedppds = drivertypes.get_ordered_ppdnames (orderedtypes,
                                                                ppdsdict, fit)
                print(mm + ":")
                i = 1
                for t, ppd in orderedppds:
                    print("%d  %s\n    (%s)" % (i, ppd, t))
                    i += 1

                print()



#!/usr/bin/python3

## system-config-printer

## Copyright (C) 2008, 2009, 2010, 2011, 2012, 2013, 2014 Red Hat, Inc.
## Authors:
##  Tim Waugh <twaugh@redhat.com>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import config
import gettext
gettext.install(domain=config.PACKAGE, localedir=config.localedir)
import cups
import dbus
from gi.repository import GObject
from gi.repository import Gtk
import os
import socket
import tempfile
import time

import authconn
from debug import *
from errordialogs import *
import firewallsettings
from gui import GtkGUI

try:
    try_CUPS_SERVER_REMOTE_ANY = cups.CUPS_SERVER_REMOTE_ANY
except AttributeError:
    # cups module was compiled with CUPS < 1.3
    try_CUPS_SERVER_REMOTE_ANY = "_remote_any"

# Set up "Problems?" link button
class _UnobtrusiveButton(Gtk.Button):
    def __init__ (self, **args):
        Gtk.Button.__init__ (self, **args)
        self.set_relief (Gtk.ReliefStyle.NONE)
        label = self.get_child ()
        text = label.get_text ()
        label.set_use_markup (True)
        label.set_markup ('<span size="small" ' +
                          'underline="single" ' +
                          'color="#0000ee">%s</span>' % text)

class ServerSettings(GtkGUI):

    __gsignals__ = {
        'settings-applied': (GObject.SignalFlags.RUN_LAST, None, ()),
        'dialog-canceled': (GObject.SignalFlags.RUN_LAST, None, ()),
        'problems-clicked': (GObject.SignalFlags.RUN_LAST, None, ()),
        }

    RESOURCE="/admin/conf/cupsd.conf"

    def __init__ (self, host=None, encryption=None, parent=None):
        GObject.GObject.__init__ (self)
        self.cupsconn = authconn.Connection (host=host, encryption=encryption)
        self._host = host
        self._parent = parent
        self.getWidgets({"ServerSettingsDialog":
                             ["ServerSettingsDialog",
                              "chkServerBrowse",
                              "chkServerShare",
                              "chkServerShareAny",
                              "chkServerRemoteAdmin",
                              "chkServerAllowCancelAll",
                              "chkServerLogDebug",
                              "hboxServerBrowse",
                              "rbPreserveJobFiles",
                              "rbPreserveJobHistory",
                              "rbPreserveJobNone",
                              "tvBrowseServers",
                              "frameBrowseServers",
                              "btAdvServerAdd",
                              "btAdvServerRemove"]},

                        domain=config.PACKAGE)

        problems = _UnobtrusiveButton (label=_("Problems?"))
        self.hboxServerBrowse.pack_end (problems, False, False, 0)
        problems.connect ('clicked', self.problems_clicked)
        problems.show ()

        self.ServerSettingsDialog.connect ('response', self.on_response)

        # Signal handler IDs.
        self.handler_ids = {}

        self.dialog = self.ServerSettingsDialog
        self.browse_treeview = self.tvBrowseServers
        self.add = self.btAdvServerAdd
        self.remove = self.btAdvServerRemove

        selection = self.browse_treeview.get_selection ()
        selection.set_mode (Gtk.SelectionMode.MULTIPLE)
        self._connect (selection, 'changed', self.on_treeview_selection_changed)

        for column in self.browse_treeview.get_columns():
            self.browse_treeview.remove_column(column)
        col = Gtk.TreeViewColumn ('', Gtk.CellRendererText (), text=0)
        self.browse_treeview.append_column (col)

        self._fillAdvanced ()
        self._fillBasic ()

        if parent:
            self.dialog.set_transient_for (parent)

        self.connect_signals ()
        self.dialog.show ()

    def get_dialog (self):
        return self.dialog

    def problems_clicked (self, button):
        self.emit ('problems-clicked')

    def _fillAdvanced(self):
        # Fetch cupsd.conf
        f = tempfile.TemporaryFile () # mode='w+b'
        try:
            self.cupsconn.getFile (self.RESOURCE, file=f)
        except cups.HTTPError as e:
            (s,) = e.args
            show_HTTP_Error (s, self._parent)
            raise

        def parse_yesno (line):
            arg1 = line.split (' ')[1].strip ()
            if arg1 in ['true', 'on', 'enabled', 'yes']:
                return True
            if arg1 in ['false', 'off', 'disabled', 'no', '0']:
                return False
            try:
                if int (arg1) != 0:
                    return True
            except:
                pass
            raise RuntimeError

        preserve_job_history = True
        preserve_job_files = True
        browsing = True
        self.browse_poll = []
        f.seek (0)
        for line in f:
            line = line.decode ('UTF-8')
            l = line.lower ().strip ()
            if l.startswith ("preservejobhistory "):
                try:
                    preserve_job_history = parse_yesno (l)
                except:
                    pass
            elif l.startswith ("preservejobfiles "):
                try:
                    preserve_job_files = parse_yesno (l)
                except:
                    pass
            elif l.startswith ("browsing "):
                try:
                    browsing = parse_yesno (l)
                except:
                    pass
            elif l.startswith ("browsepoll "):
                self.browse_poll.append (line[len ("browsepoll "):].strip ())

        self.frameBrowseServers.set_sensitive (browsing)

        if preserve_job_files:
            self.rbPreserveJobFiles.set_active (True)
        elif preserve_job_history:
            self.rbPreserveJobHistory.set_active (True)
        else:
            self.rbPreserveJobNone.set_active (True)

        self.preserve_job_history = preserve_job_history
        self.preserve_job_files = preserve_job_files

        model = Gtk.ListStore (str)
        self.browse_treeview.set_model (model)
        for server in self.browse_poll:
            model.append (row=[server])

    def _fillBasic(self):
        self.changed = set()
        self.cupsconn._begin_operation (_("fetching server settings"))
        try:
            self.server_settings = self.cupsconn.adminGetServerSettings()
        except cups.IPPError as e:
            (e, m) = e.args
            show_IPP_Error(e, m, self._parent)
            self.cupsconn._end_operation ()
            raise

        self.cupsconn._end_operation ()

        for widget, setting in [
            (self.chkServerBrowse, cups.CUPS_SERVER_REMOTE_PRINTERS),
            (self.chkServerShare, cups.CUPS_SERVER_SHARE_PRINTERS),
            (self.chkServerShareAny, try_CUPS_SERVER_REMOTE_ANY),
            (self.chkServerRemoteAdmin, cups.CUPS_SERVER_REMOTE_ADMIN),
            (self.chkServerAllowCancelAll, cups.CUPS_SERVER_USER_CANCEL_ANY),
            (self.chkServerLogDebug, cups.CUPS_SERVER_DEBUG_LOGGING),]:
            widget.setting = setting
            if setting in self.server_settings:
                widget.set_active(int(self.server_settings[setting]))
                widget.set_sensitive(True)
                widget.show()
            else:
                widget.set_active(False)
                widget.set_sensitive(False)
                widget.hide()

        if cups.CUPS_SERVER_REMOTE_PRINTERS in self.server_settings:
            self.frameBrowseServers.show()
        else:
            self.frameBrowseServers.hide()

        try:
            flag = cups.CUPS_SERVER_SHARE_PRINTERS
            publishing = int (self.server_settings[flag])
            self.server_is_publishing = publishing
        except AttributeError:
            pass

        # Set sensitivity of 'Allow printing from the Internet'.
        self.on_server_changed (self.chkServerShare) # (any will do here)

    def on_server_changed(self, widget):
        debugprint ("on_server_changed: %s" % widget)
        setting = widget.setting
        if setting in self.server_settings:
            if str(int(widget.get_active())) == self.server_settings[setting]:
                self.changed.discard(widget)
            else:
                self.changed.add(widget)

        sharing = self.chkServerShare.get_active ()
        self.chkServerShareAny.set_sensitive (
            sharing and try_CUPS_SERVER_REMOTE_ANY in self.server_settings)

    def _connect (self, widget, signal, handler, reason=None):
        id = widget.connect (signal, handler)
        if reason not in self.handler_ids:
            self.handler_ids[reason] = []
        self.handler_ids[reason].append ((widget, id))

    def _disconnect (self, reason=None):
        if reason in self.handler_ids:
            for (widget, id) in self.handler_ids[reason]:
                widget.disconnect (id)
            del self.handler_ids[reason]

    def on_treeview_selection_changed (self, selection):
        self.remove.set_sensitive (selection.count_selected_rows () != 0)

    def on_add_clicked (self, button):
        model = self.browse_treeview.get_model ()
        iter = model.insert (0, row=[_("Enter hostname")])
        button.set_sensitive (False)
        col = self.browse_treeview.get_columns ()[0]
        cell = col.get_cells ()[0]
        cell.set_property ('editable', True)
        self.browse_treeview.set_cursor (Gtk.TreePath(), col, True)
        self._connect (cell, 'edited', self.on_browse_poll_edited, 'edit')
        self._connect (cell, 'editing-canceled',
                       self.on_browse_poll_edit_cancel, 'edit')

    def on_browse_poll_edited (self, cell, path, newvalue):
        model = self.browse_treeview.get_model ()
        iter = model.get_iter (path)
        model.set_value (iter, 0, newvalue)
        cell.stop_editing (False)
        cell.set_property ('editable', False)
        self.add.set_sensitive (True)
        self._disconnect ('edit')

        valid = True
        # Check that it's a valid IP address or hostname.
        # First, is it an IP address?
        try:
            socket.getaddrinfo (newvalue, '0', socket.AF_UNSPEC, 0, 0,
                                socket.AI_NUMERICHOST)
        except socket.gaierror:
            # No.  Perhaps it's a hostname.
            labels = newvalue.split (".")
            seen_alpha = False
            for label in labels:
                if (label[0] == '-' or
                    label.endswith ('-')):
                    valid = False
                    break
                for char in label:
                    if not seen_alpha:
                        if char.isalpha ():
                            seen_alpha = True

                    if not (char.isalpha () or
                            char.isdigit () or
                            char == '-'):
                        valid = False
                        break

                if not valid:
                    break

            if valid and not seen_alpha:
                valid = False

        if valid:
            count = 0
            i = model.get_iter_first ()
            while i:
                if model.get_value (i, 0) == newvalue:
                    count += 1
                    if count == 2:
                        valid = False
                        selection = self.browse_treeview.get_selection ()
                        selection.select_iter (i)
                        break
                i = model.iter_next (i)
        else:
            model.remove (iter)

    def on_browse_poll_edit_cancel (self, cell):
        cell.stop_editing (True)
        cell.set_property ('editable', False)
        model = self.browse_treeview.get_model ()
        iter = model.get_iter (Gtk.TreePath())
        model.remove (iter)
        self.add.set_sensitive (True)
        self.remove.set_sensitive (False)
        self._disconnect ('edit')

    def on_remove_clicked (self, button):
        model = self.browse_treeview.get_model ()
        selection = self.browse_treeview.get_selection ()
        rows = selection.get_selected_rows ()
        refs = [Gtk.TreeRowReference.new (model, path) for path in rows[1]]
        for ref in refs:
            path = ref.get_path ()
            iter = model.get_iter (path)
            model.remove (iter)

    def on_response (self, dialog, response):
        if (response == Gtk.ResponseType.CANCEL or
            response != Gtk.ResponseType.OK):
            self._disconnect ()
            self.dialog.hide ()
            self.emit ('dialog-canceled')
            del self
            return

        self.saveBasic ()
        self.saveAdvanced ()

    def _reconnect (self):
        # Now reconnect, in case the server needed to reload.
        try:
            attempt = 1
            while attempt <= 5:
                try:
                    self.cupsconn._connect ()
                    break
                except RuntimeError:
                    # Connection failed.
                    time.sleep (1)
                    attempt += 1
        except AttributeError:
            # _connect method is part of the authconn.Connection
            # interface, so don't fail if that method doesn't exist.
            pass

    def saveAdvanced (self):
        # See if there are changes.
        preserve_job_files = self.rbPreserveJobFiles.get_active ()
        preserve_job_history = (preserve_job_files or
                                self.rbPreserveJobHistory.get_active ())
        model = self.browse_treeview.get_model ()
        browse_poll = []
        iter = model.get_iter_first ()
        while iter:
            browse_poll.append (model.get_value (iter, 0))
            iter = model.iter_next (iter)

        if (set (browse_poll) == set (self.browse_poll) and
            preserve_job_files == self.preserve_job_files and
            preserve_job_history == self.preserve_job_history):
            self._disconnect ()
            self.dialog.hide ()
            self.emit ('settings-applied')
            del self
            return

        # Fetch cupsd.conf afresh
        f = tempfile.TemporaryFile () # mode='w+b'
        try:
            self.cupsconn.getFile (self.RESOURCE, file=f)
        except cups.HTTPError as e:
            (s,) = e.args
            show_HTTP_Error (s, self.dialog)
            return

        job_history_line = job_files_line = browsepoll_lines = ""

        # Default is to preserve job history
        if not preserve_job_history:
            job_history_line = "PreserveJobHistory No\n"

        # Default is to preserve job files.
        if not preserve_job_files:
            job_files_line = "PreserveJobFiles No\n"

        for server in browse_poll:
            browsepoll_lines += "BrowsePoll %s\n" % server

        f.seek (0)
        conf = tempfile.TemporaryFile () # mode='w+b'
        wrote_preserve_history = wrote_preserve_files = False
        wrote_browsepoll = False
        has_browsepoll = False
        for line in f:
            line = line.decode('UTF-8')
            l = line.lower ().strip ()
            if l.startswith ("browsepoll "):
                has_browsepoll = True
                break

        # Return to the start of file
        f.seek(0)

        for line in f:
            line = line.decode('UTF-8')
            l = line.lower ().strip ()
            if l.startswith ("preservejobhistory "):
                if wrote_preserve_history:
                    # Don't write out another line with this keyword.
                    continue
                # Alter this line before writing it out.
                line = job_history_line
                wrote_preserve_history = True
            elif l.startswith ("preservejobfiles "):
                if wrote_preserve_files:
                    # Don't write out another line with this keyword.
                    continue
                # Alter this line before writing it out.
                line = job_files_line
                wrote_preserve_files = True
            elif (has_browsepoll and
                  l.startswith ("browsepoll ")):
                if wrote_browsepoll:
                    # Ignore extra BrowsePoll lines.
                    continue
                # Write new BrowsePoll section.
                conf.write (browsepoll_lines.encode('UTF-8'))
                wrote_browsepoll = True
                # Don't write out the original BrowsePoll line.
                continue
            elif (not has_browsepoll and
                  l.startswith ("browsing ")):
                if not wrote_browsepoll:
                    # Write original Browsing line.
                    conf.write (line.encode('UTF-8'))
                    # Write new BrowsePoll section.
                    conf.write (browsepoll_lines.encode('UTF-8'))
                    wrote_browsepoll = True
                    continue

            conf.write (line.encode('UTF-8'))

        if not wrote_preserve_history:
            conf.write (job_history_line.encode('UTF-8'))
        if not wrote_preserve_files:
            conf.write (job_files_line.encode('UTF-8'))
        if not wrote_browsepoll:
            conf.write (browsepoll_lines.encode('UTF-8'))

        conf.flush ()
        fd = conf.fileno ()
        os.lseek (fd, 0, os.SEEK_SET)
        try:
            self.cupsconn.putFile ("/admin/conf/cupsd.conf", fd=fd)
        except cups.IPPError as e:
            (e, m) = e.args
            show_IPP_Error (e, m, self.dialog)
            return
        except cups.HTTPError as e:
            (s,) = e.args
            show_HTTP_Error (s, self.dialog)
            return

        # Give the server a chance to process our request.
        time.sleep (1)

        self._reconnect ()

        self._disconnect ()
        self.emit ('settings-applied')
        self.dialog.hide ()
        del self

    def saveBasic (self):
        setting_dict = dict()
        for widget, setting in [
            (self.chkServerBrowse, cups.CUPS_SERVER_REMOTE_PRINTERS),
            (self.chkServerShare, cups.CUPS_SERVER_SHARE_PRINTERS),
            (self.chkServerShareAny, try_CUPS_SERVER_REMOTE_ANY),
            (self.chkServerRemoteAdmin, cups.CUPS_SERVER_REMOTE_ADMIN),
            (self.chkServerAllowCancelAll, cups.CUPS_SERVER_USER_CANCEL_ANY),
            (self.chkServerLogDebug, cups.CUPS_SERVER_DEBUG_LOGGING),]:
            if setting not in self.server_settings: continue
            setting_dict[setting] = str(int(widget.get_active()))
        self.cupsconn._begin_operation (_("modifying server settings"))
        try:
            self.cupsconn.adminSetServerSettings(setting_dict)
        except cups.IPPError as e:
            (e, m) = e.args
            show_IPP_Error(e, m, self.dialog)
            self.cupsconn._end_operation ()
            return True
        except RuntimeError as s:
            show_IPP_Error(None, s, self.dialog)
            self.cupsconn._end_operation ()
            return True
        self.cupsconn._end_operation ()
        self.changed = set()

        old_setting = self.server_settings.get (cups.CUPS_SERVER_SHARE_PRINTERS,
                                                '0')
        new_setting = setting_dict.get (cups.CUPS_SERVER_SHARE_PRINTERS, '0')
        if (old_setting == '0' and new_setting != '0'):
            # We have just enabled print queue sharing.
            # Let's see if the firewall will allow IPP TCP packets in.
            try:
                if (self._host == 'localhost' or
                    self._host[0] == '/'):
                    f = firewallsettings.FirewallD ()
                    if not f.running:
                        f = firewallsettings.SystemConfigFirewall ()

                    allowed = f.check_ipp_server_allowed ()
                else:
                    # This is a remote server.  Nothing we can do
                    # about the firewall there.
                    allowed = True

                if not allowed:
                    dialog = Gtk.MessageDialog (parent=self.ServerSettingsDialog,
                                                modal=True, destroy_with_parent=True,
                                                message_type=Gtk.MessageType.QUESTION,
                                                buttons=Gtk.ButtonsType.NONE,
                                                text=_("Adjust Firewall"))
                    dialog.format_secondary_text (_("Adjust the firewall now "
                                                    "to allow all incoming IPP "
                                                    "connections?"))
                    dialog.add_buttons (Gtk.STOCK_CANCEL, Gtk.ResponseType.NO,
                                        _("Adjust Firewall"), Gtk.ResponseType.YES)
                    response = dialog.run ()
                    dialog.destroy ()

                    if response == Gtk.ResponseType.YES:
                        f.add_service (firewallsettings.IPP_SERVER_SERVICE)
                        f.write ()
            except (dbus.DBusException, Exception):
                nonfatalException ()

        time.sleep(1) # give the server a chance to process our request

        # Now reconnect, in case the server needed to reload.
        self._reconnect ()

if __name__ == '__main__':
    os.environ['SYSTEM_CONFIG_PRINTER_UI'] = 'ui'
    loop = GObject.MainLoop ()

    def quit (*args):
        loop.quit ()

    def problems (obj):
        print("%s: problems" % obj)

    set_debugging (True)
    s = ServerSettings ()
    s.connect ('dialog-canceled', quit)
    s.connect ('settings-applied', quit)
    s.connect ('problems-clicked', problems)
    loop.run ()



#!/usr/bin/python3

## system-config-printer

## Copyright (C) 2010, 2011, 2012, 2013, 2014 Red Hat, Inc.
## Author: Tim Waugh <twaugh@redhat.com>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import dbus
from gi.repository import GObject
from gi.repository import Gtk
import cupshelpers

import cups
cups.require ("1.9.52")

import asyncconn
from debug import debugprint
import config
import gettext
gettext.install(domain=config.PACKAGE, localedir=config.localedir)

class PPDsLoader(GObject.GObject):
    """
    1. If PackageKit support is available, and this is a local server,
    try to use PackageKit to install relevant drivers.  We do this
    because we can only make the right choice about the "best" driver
    when the full complement of drivers is there to choose from.

    2. Fetch the list of available drivers from CUPS.

    3. If Jockey is available, and there is no appropriate driver
    available, try to use Jockey to install one.

    4. If Jockey was able to install one, fetch the list of available
    drivers again.
    """

    __gsignals__ = {
        'finished': (GObject.SignalFlags.RUN_LAST, None, ())
        }

    def __init__ (self, device_id=None, parent=None, device_uri=None,
                  host=None, encryption=None, language=None,
                  device_make_and_model=None):
        GObject.GObject.__init__ (self)
        debugprint ("+%s" % self)
        self._device_id = device_id
        self._device_uri = device_uri
        self._device_make_and_model = device_make_and_model
        self._parent = parent
        self._host = host
        self._encryption = encryption
        self._language = language

        self._installed_files = []
        self._conn = None
        self._ppds = None
        self._exc = None

        self._ppdsmatch_result = None
        self._jockey_queried = False
        self._jockey_has_answered = False
        self._local_cups = (self._host is None or
                            self._host == "localhost" or
                            self._host[0] == '/')
        try:
            self._bus = dbus.SessionBus ()
        except:
            debugprint ("Failed to get session bus")
            self._bus = None

        fmt = _("Searching")
        self._dialog = Gtk.MessageDialog (parent=parent,
                                          modal=True, destroy_with_parent=True,
                                          message_type=Gtk.MessageType.INFO,
                                          buttons=Gtk.ButtonsType.CANCEL,
                                          text=fmt)

        self._dialog.format_secondary_text (_("Searching for drivers"))

        self._dialog.connect ("response", self._dialog_response)

    def run (self):
        self._dialog.show_all ()

        if self._device_id:
            self._devid_dict = cupshelpers.parseDeviceID (self._device_id)

        if (self._local_cups and
            self._device_id and
            self._devid_dict["MFG"] and
            self._devid_dict["MDL"] and
            self._bus):
            self._gpk_device_id = "MFG:%s;MDL:%s;" % (self._devid_dict["MFG"],
                                                      self._devid_dict["MDL"])
            self._query_packagekit ()
        else:
            self._query_cups ()

    def __del__ (self):
        debugprint ("-%s" % self)

    def destroy (self):
        debugprint ("DESTROY: %s" % self)
        if self._dialog:
            self._dialog.destroy ()
            self._dialog = None

        self._parent = None

        if self._conn:
            self._conn.destroy ()
            self._conn = None

    def get_installed_files (self):
        return self._installed_files

    def get_ppds (self):
        return self._ppds

    def get_ppdsmatch_result (self):
        return self._ppdsmatch_result

    def get_error (self):
        debugprint ("%s: stored error is %s" % (self, repr (self._exc)))
        return self._exc

    def _dialog_response (self, dialog, response):
        dialog.destroy ()
        self._dialog = None
        self.emit ('finished')

    def _query_cups (self):
        debugprint ("Asking CUPS for PPDs")
        if (not self._conn):
            c = asyncconn.Connection (host=self._host,
                                      encryption=self._encryption,
                                      reply_handler=self._cups_connect_reply,
                                      error_handler=self._cups_error)
            self._conn = c
        else:
            self._cups_connect_reply(self._conn, None)

    def _cups_connect_reply (self, conn, UNUSED):
        conn._begin_operation (_("fetching PPDs"))
        conn.getPPDs2 (reply_handler=self._cups_reply,
                       error_handler=self._cups_error)

    def _cups_reply (self, conn, result):
        ppds = cupshelpers.ppds.PPDs (result, language=self._language)
        self._ppds = ppds
        self._need_requery_cups = False
        if self._device_id:
            fit = ppds.\
                getPPDNamesFromDeviceID (self._devid_dict["MFG"],
                                         self._devid_dict["MDL"],
                                         self._devid_dict["DES"],
                                         self._devid_dict["CMD"],
                                         self._device_uri,
                                         self._device_make_and_model)

            ppdnamelist = ppds.\
                orderPPDNamesByPreference (list(fit.keys ()),
                                           self._installed_files,
                                           devid=self._devid_dict,
                                           fit=fit)
            self._ppdsmatch_result = (fit, ppdnamelist)

            ppdname = ppdnamelist[0]
            if (self._bus and
                not fit[ppdname].startswith ("exact") and
                not self._jockey_queried and
                self._devid_dict["MFG"] and
                self._devid_dict["MDL"] and
                self._local_cups):
                # Try to install packages using jockey if
                # - there's no appropriate driver (PPD) locally available
                # - we are configuring local CUPS server
                self._jockey_queried = True
                self._query_jockey ()
                return

        conn.destroy ()
        self._conn = None
        if self._dialog is not None:
            self._dialog.destroy ()
            self._dialog = None

        self.emit ('finished')

    def _cups_error (self, conn, exc):
        conn.destroy ()
        self._conn = None
        self._ppds = None
        self._exc = exc
        if self._dialog is not None:
            self._dialog.destroy ()
            self._dialog = None

        self.emit ('finished')

    def _query_packagekit (self):
        debugprint ("Asking PackageKit to install drivers")
        try:
            obj = self._bus.get_object ("org.freedesktop.PackageKit",
                                        "/org/freedesktop/PackageKit")
            proxy = dbus.Interface (obj, "org.freedesktop.PackageKit.Modify")
            resources = [self._gpk_device_id]
            interaction = "hide-finished"
            debugprint ("Calling InstallPrinterDrivers (%s, %s, %s)" %
                        (repr (0), repr (resources), repr (interaction)))
            proxy.InstallPrinterDrivers (dbus.UInt32 (0),
                                         resources, interaction,
                                         reply_handler=self._packagekit_reply,
                                         error_handler=self._packagekit_error,
                                         timeout=3600)
        except Exception as e:
            debugprint ("Failed to talk to PackageKit: %s" % repr (e))
            if self._dialog:
                self._dialog.show_all ()
                self._query_cups ()

    def _packagekit_reply (self):
        debugprint ("Got PackageKit reply")
        self._need_requery_cups = True
        if self._dialog:
            self._dialog.show_all ()
            self._query_cups ()

    def _packagekit_error (self, exc):
        debugprint ("Got PackageKit error: %s" % repr (exc))
        if self._dialog:
            self._dialog.show_all ()
            self._query_cups ()

    def _query_jockey (self):
        debugprint ("Asking Jockey to install drivers")
        try:
            obj = self._bus.get_object ("com.ubuntu.DeviceDriver", "/GUI")
            jockey = dbus.Interface (obj, "com.ubuntu.DeviceDriver")
            r = jockey.search_driver ("printer_deviceid:%s" % self._device_id,
                                      reply_handler=self._jockey_reply,
                                      error_handler=self._jockey_error,
                                      timeout=3600)
        except Exception as e:
            self._jockey_error (e)

    def _jockey_reply (self, conn, result):
        debugprint ("Got Jockey result: %s" % repr (result))
        self._jockey_has_answered = True
        try:
            self._installed_files = result[1]
        except:
            self._installed_files = []
        self._query_cups ()

    def _jockey_error (self, exc):
        debugprint ("Got Jockey error: %s" % repr (exc))
        if self._need_requery_cups:
            self._query_cups ()
        else:
            if self._conn is not None:
                self._conn.destroy ()
                self._conn = None

            if self._dialog is not None:
                self._dialog.destroy ()
                self._dialog = None

            self.emit ('finished')

if __name__ == "__main__":
    class Foo:
        def __init__ (self):
            w = Gtk.Window ()
            b = Gtk.Button.new_with_label ("Go")
            w.add (b)
            b.connect ('clicked', self.go)
            w.connect ('delete-event', Gtk.main_quit)
            w.show_all ()
            self._window = w

        def go (self, button):
            loader = PPDsLoader (device_id="MFG:MFG;MDL:MDL;",
                                 parent=self._window)
            loader.connect ('finished', self.ppds_loaded)
            loader.run ()

        def ppds_loaded (self, ppdsloader):
            self._window.destroy ()
            Gtk.main_quit ()
            exc = ppdsloader.get_error ()
            print(exc)
            ppds = ppdsloader.get_ppds ()
            if ppds is not None:
                print(len (ppds))

            ppdsloader.destroy ()

    from debug import set_debugging
    set_debugging (True)
    Foo ()
    Gtk.main ()
